<?php

namespace Drupal\github_connect\Controller;

use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Database\Connection;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Routing\UrlGeneratorInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\github_connect\GithubConnectService;
use Drupal\user\UserInterface;
use Drupal\user\UserStorageInterface;
use GuzzleHttp\ClientInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 * Class GithubConnectController.
 *
 * @package Drupal\github_connect\Controller
 */
class GithubConnectController extends ControllerBase implements ContainerInjectionInterface {

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected AccountInterface $account;

  /**
   * The URL of the link.
   *
   * @var \Drupal\Core\Routing\UrlGeneratorInterface
   */
  protected UrlGeneratorInterface $url;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $connection;

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\Config
   */
  protected Config $config;

  /**
   * The HTTP client to fetch the feed data with.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected ClientInterface $httpClient;

  /**
   * GithubController Class Service.
   *
   * @var \Drupal\github_connect\GithubConnectService
   */
  protected GithubConnectService $githubConnectService;

  /**
   * The current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected Request $request;

  protected UserStorageInterface $userStorage;

  /**
   * Class constructor.
   */
  public function __construct(AccountInterface $account, UrlGeneratorInterface $url, Connection $connection, ConfigFactoryInterface $config_factory, ClientInterface $client, GithubConnectService $githubConnectService, Request $request, UserStorageInterface $userStorage) {
    $this->account = $account;
    $this->url = $url;
    $this->connection = $connection;
    $this->config = $config_factory->getEditable('github_connect.settings');
    $this->httpClient = $client;
    $this->githubConnectService = $githubConnectService;
    $this->request = $request;
    $this->userStorage = $userStorage;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      // Load the service required to construct this class.
      $container->get('current_user'),
      $container->get('url_generator'),
      $container->get('database'),
      $container->get('config.factory'),
      $container->get('http_client'),
      $container->get('github_connect_service'),
      $container->get('request_stack')->getCurrentRequest(),
      $container->get('entity_type.manager')->getStorage('user')
    );
  }

  /**
   * Process an Github authentication.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse|null
   *   A redirect response, or NULL if authentication failed.
   *
   * @throws \GuzzleHttp\Exception\GuzzleException
   */
  public function githubConnectGetAccessToken(): ?RedirectResponse {
    // Get current user data.
    $uid = $this->account->id();
    $account = $this->account;
    $client_id = $this->config->get('client_id');
    $client_secret = $this->config->get('client_secret');

    // Assemble the URL to get the access token @Github.
    $github_code = $this->request->query->get('code');
    $github_base_url = 'https://github.com/login/oauth/access_token?';
    $github_parameters = "client_id=$client_id&client_secret=$client_secret&code=$github_code";
    $url = $github_base_url . $github_parameters;

    // Send the request to Github.
    $client = $this->httpClient;
    $response = $client->request('POST', $url);
    $body = $response->getBody();

    // Get the actual access token
    parse_str($body, $tokens);
    $access_token = $tokens['access_token'] ?? NULL;

    if (empty($access_token)) {
      return NULL;
    }

    if ($access_token) {
      // Check if a user exists for the token.
      // Get user details from GitHub to handle user association.
      $github_user = $this->githubConnectService->githubConnectGetGithubUserInfo($access_token);
      if ($github_user && !empty($github_user['html_url'])) {
        // Check the authmap for an existing associated account.
        $account = $this->githubConnectUserExternalLoad($github_user['html_url']);
      }

      if ($uid == 0) {
        // First the case where an anonymous user attempts a login.
        if ($account) {
          // If there is a user with the token log that user in.
          $this->githubConnectService->githubConnectUserLogin($account);
          return $this->redirect('<front>');
        }
        else {
          // Otherwise register the user and log in.
          $github_user = $this->githubConnectService->githubConnectGetGithubUserInfo($access_token);

          if ($existing_user_by_mail = user_load_by_mail($github_user['email'])) {
            // If a user with this email address exists, let him connect the
            // github account to his already created account.
            return $this->redirect('github_connect.verify', ['uid' => $existing_user_by_mail->id(), 'token' => $access_token]);
          }
          else {
            // Otherwise make sure there is no account with the same username.
            if ($existing_user_by_name = user_load_by_name($github_user['login'])) {
              return $this->redirect('github.username', ['user' => $existing_user_by_name->id(), 'token' => $access_token]);
            }
            else {
              $this->githubConnectService->githubConnectRegister($github_user, $access_token);
              return $this->redirect('<front>');
            }
          }
        }
      }
      else {
        // Second the case where an logged in user attempts to attach his github
        // account.
        if ($account) {
          // If there is a user with the token, throw an error.
          $this->messenger()->addError($this->t('Your GitHub account could not be connected, it is already coupled with another user.'));
          return $this->redirect('entity.user.canonical', ['user' => $account->id()]);
        }
        else {
          $github_user = $this->githubConnectService->githubConnectGetGithubUserInfo($access_token);

          if (!$github_user['email']) {
            $this->messenger()->addError($this->t('We could not connect your GitHub account. You need to have a public email address registered with your GitHub account.'));
            return $this->redirect('entity.user.canonical', ['user' => $uid]);
          }

          if ($github_user['html_url']) {
            $this->githubConnectService->githubConnectSaveGithubUser($account, $access_token);
            $this->messenger()->addStatus($this->t('Your GitHub account is now connected.'));
            return $this->redirect('entity.user.canonical', ['user' => $uid]);
          }
        }
      }
    }
    else {
      // If we didn't get a token, connection to Github failed.
      $this->messenger()->addError($this->t('Failed connecting to GitHub.'));
      return $this->redirect('<front>');
    }
    return NULL;
  }

  /**
   * Get user from GitHub access token.
   *
   * @param string $token
   *   Access token from GitHub.
   *
   * @return \Drupal\user\UserInterface|null
   *   The Drupal user, or NULL if not found.
   */
  public function githubConnectGetTokenUser(string $token): ?UserInterface {
    if ($token) {
      $uid = $this->connection->select('github_connect_users', 'g_u')
        ->fields('g_u', ['uid'])
        ->condition('access_token', $token)
        ->execute()
        ->fetchField();
      return empty($uid) ? NULL : $this->userStorage->load($uid);
    }
    return NULL;
  }

  /**
   * Authorizes whether correct user is logged in or not.
   *
   * @param string $authname
   *   The authname to match.
   *
   * @return \Drupal\user\UserInterface|null
   *   The Drupal user, or NULL if not found.
   */
  public function githubConnectUserExternalLoad(string $authname): ?UserInterface {
    $uid = $this->connection->select('github_connect_authmap', 'gca')
      ->fields('gca', ['uid'])
      ->condition('authname', $authname)
      ->execute()->fetchField();

    return empty($uid) ? NULL : $this->userStorage->load($uid);
  }

}
