<?php

namespace Drupal\github_connect;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Password\PasswordGeneratorInterface;
use Drupal\Core\Url;
use Drupal\user\UserInterface;
use Drupal\user\UserStorageInterface;
use GuzzleHttp\ClientInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * WeatherService.
 */
class GithubConnectService {

  use MessengerTrait;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $database;

  /**
   * The HTTP client to fetch the feed data with.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected ClientInterface $httpClient;

  /**
   * The HTTP request time as a Unix timestamp.
   *
   * @var int
   */
  protected int $requestTime;

  /**
   * The password generator service.
   *
   * @var \Drupal\Core\Password\PasswordGeneratorInterface
   */
  protected PasswordGeneratorInterface $passwordGenerator;

  /**
   * The user storage service.
   *
   * @var \Drupal\user\UserStorageInterface
   */
  protected UserStorageInterface $userStorage;

  /**
   * Class constructor.
   */
  public function __construct(Connection $database, ClientInterface $client, TimeInterface $time, PasswordGeneratorInterface $passwordGenerator, EntityTypeManagerInterface $entityTypeManager) {
    $this->database = $database;
    $this->httpClient = $client;
    $this->requestTime = $time->getRequestTime();
    $this->passwordGenerator = $passwordGenerator;
    $this->userStorage = $entityTypeManager->getStorage('user');
  }

  /**
   * Saves the new GitHub user in github_connect_users.
   *
   * @param \Drupal\user\UserInterface $account
   *   The user to be saved.
   * @param string $token
   *   The user's auth token.
   */
  public function githubConnectSaveGithubUser(UserInterface $account, string $token): void {
    $github_user = self::githubConnectGetGithubUserInfo($token);

    $this->database->insert('github_connect_authmap')
      ->fields([
        'uid' => $account->id(),
        'provider' => 'github_connect',
        'authname' => $github_user['html_url'],
      ])
      ->execute();

    // Store GitHub user with token.
    $this->database->insert('github_connect_users')
      ->fields([
        'uid' => $account->id(),
        'access_token' => $token,
        'timestamp' => $this->requestTime,
      ])
      ->execute();
  }

  /**
   * Gets the user info provided by github.
   *
   * @param string $token
   *   The token for the github user.
   *
   * @return array
   *   User info as returned from Github.
   */
  public function githubConnectGetGithubUserInfo(string $token): array {
    $cache = &drupal_static(__FUNCTION__);

    if (!is_null($cache)) {
      $github_user = $cache;
    }
    else {
      // Collects the User information from GitHub.
      $client = $this->httpClient;
      $ghuser = $client->request('GET', 'https://api.github.com/user?access_token=' . $token . '&scope=user&token_type=bearer');
      // TODO pass timeout value.
      $data = (string) $ghuser->getBody();
      $github_user = Json::decode($data);
      $github_user_emails = self::githubConnectGetGithubUserEmails($token);
      $github_user['email'] = $github_user_emails[0]['email'];
    }

    return $github_user;
  }

  /**
   * Get the private email addresses from the user.
   *
   * @param string $token
   *   The Github access token for the user.
   *
   * @return array
   *   A structured array of email info.
   */
  public function githubConnectGetGithubUserEmails(string $token): array {
    $cache = &drupal_static(__FUNCTION__);

    if (!is_null($cache)) {
      $github_user_emails = $cache;
    }
    else {
      // Collects the User information from GitHub.
      $client = $this->httpClient;
      $ghuser = $client->request('GET', 'https://api.github.com/user/emails?access_token=' . $token . '&scope=user&token_type=bearer');
      $data = (string) $ghuser->getBody();
      $github_user_emails = Json::decode($data);
    }

    return $github_user_emails;
  }

  /**
   * Register new user.
   *
   * @param array $github_user
   *   The user info returned from github.
   * @param string $token
   *   Github's access token for this user.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse|null
   */
  public function githubConnectRegister(array $github_user, string $token): ?RedirectResponse {
    $username = $github_user['login'];

    $userinfo = [
      'name' => $username,
      'mail' => $github_user['email'],
      'pass' => $this->passwordGenerator->generate(),
      'status' => 1,
      'access' => $this->requestTime,
      'init' => $github_user['email'],
    ];

    // Import GitHub profile picture on first login.
    if (empty(user_load_by_mail($github_user['email'])) && !empty($github_user['avatar_url'])) {
      // Download profile image from Github.
      $avatar_url = file_get_contents($github_user['avatar_url']);
      $file = file_save_data($avatar_url, 'public://' . time() . '_avatars.png', NULL);
      $userinfo['user_picture'] = $file->id();
    }

    $account = $this->userStorage->create($userinfo);
    try {
      $account->save();
      $this->githubConnectSaveGithubUser($account, $token);

      // Log in the stored user.
      self::githubConnectUserLogin($account);
      return new RedirectResponse(Url::fromRoute('<front>')->toString());
    }
    catch (\Exception $e) {
      $this->messenger()->addError($this->t('Error saving new user.'));
    }

  }

  /**
   * Log the user with the given account in.
   */
  public function githubConnectUserLogin($account) {
    $uid = $account->id();
    $user = $this->userStorage->load($uid);
    user_login_finalize($user);
  }

}
